import collections
import itertools

import numpy as np

__all__ = [
    'isconst',
    'meshshape',
    'meshndim',
    'meshgrid',
    'inv_perm',
    'unique_rows',
    'intersect_rows',
    'repeat',
]

def isconst(x, axis=None, **kwargs):
    x = np.asanyarray(x)
    
    if axis is None:
        x = x.reshape(-1)
    else:
        if isinstance(axis, int):
            axis = [axis]
        if len(axis) == 0:
            return np.ones(x.shape, dtype=bool)
        axis = sorted([d % x.ndim for d in axis])[::-1]
        for d in axis:
            x = np.moveaxis(x, d,-1)
        x = x.reshape(*x.shape[:-len(axis)],-1)
        
    if isinstance(x, np.floating):
        return np.isclose(x[...,:-1], x[...,1:], **kwargs).all(axis=-1)
    return (x[...,:-1] == x[...,1:]).all(axis=-1)

def meshshape(*shapes, indexing='ij'):
    """
    The fundamental step of meshgrid, which is useful on its own as a standalone
    function. Given n shapes
    (s^1_1, ..., s^1_{D_1}), (s^2_1, ..., s^2_{D_2}), ..., (s^n_1, ..., s^n_{D_n}),
    Yields shapes
    (s^1_1, ..., s^1_{D_1}, 1, ..............., 1, ..., 1, ..............., 1),
    (1, ..............., 1, s^2_1, ..., s^2_{D_2}, ..., 1, ..............., 1),
    ...
    (1, ..............., 1, 1, ..............., 1, ..., s^n_1, ..., s^n_{D_n}),
    If indexing == 'xy', yields shapes
    (1, ..............., 1, s^1_1, ..., s^1_{D_1}, ..., 1, ..............., 1),
    (s^2_1, ..., s^2_{D_2}, 1, ..............., 1, ..., 1, ..............., 1),
    ...
    (1, ..............., 1, 1, ..............., 1, ..., s^n_1, ..., s^n_{D_n}),
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")

    if indexing == 'xy':
        if len(shapes) < 2:
            raise ValueError(f"Must have at least 2 shapes if indexing == 'xy', but got {len(shapes)=}.")
        
        shapes = (shapes[1], shapes[0], *shapes[2:])
        out = list(meshshape(*shapes, indexing='ij'))
        out[1], out[0] = out[0], out[1]
        yield from out
    
    sumndim = sum(len(shape) for shape in shapes)
    cumndim = 0
    for shape in shapes:
        ndim = len(shape)
        yield (1,) * cumndim + shape + (1,) * (sumndim - cumndim - ndim)
        cumndim += ndim

def meshndim(*ndims, indexing='ij'):
    """
    Same as meshshape, but yields indices instead.
    Note that reshaping using indices is faster than reshaping using shapes in numpy,
    but the opposite is true in pytorch.
    
    f1 = lambda a, n, m: a.reshape((1,) * n  + a.shape + (1,) * m)
    f2 = lambda a, n, m: a[(None,) * n + (slice(None),) * a.ndim + (None,) * m]
    shape, n, m = (3,4,100,2,6), 5, 7
    
    a = np.random.normal(shape=shape)
    %timeit f1(a, n, m)
    489 ns ± 9.55 ns per loop (mean ± std. dev. of 7 runs, 1,000,000 loops each)
    %timeit f2(a, n, m)
    472 ns ± 7.58 ns per loop (mean ± std. dev. of 7 runs, 1,000,000 loops each)

    a = torch.randn(shape)
    %timeit f1(a, n, m)
    2.62 µs ± 16.8 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)
    %timeit f2(a, n, m)
    10.8 µs ± 102 ns per loop (mean ± std. dev. of 7 runs, 100,000 loops each)
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")

    if indexing == 'xy':
        if len(shapes) < 2:
            raise ValueError(f"Must have at least 2 shapes if indexing == 'xy', but got {len(shapes)=}.")
        
        shapes = (shapes[1], shapes[0], *shapes[2:])
        out = list(meshndim(*shapes, indexing='ij'))
        out[1], out[0] = out[0], out[1]
        yield from out
    
    sumndim = sum(ndims)
    cumndim = 0
    for ndim in ndims:
        # we don't use Ellipsis because it doesn't allow for possible additional batch dimensions
        yield (None,) * cumndim + (slice(None),) * ndim + (None,) * (sumndim - cumndim - ndim)
        cumndim += ndim

def meshgrid(*tensors, indexing='ij', ndims=None):
    """
    The most general version of np.meshgrid. Meshes together tensors of shapes
    (n_1_1,...,n_1_{ndims_1},*_1), (n_2_1,...,n_2_{ndims_2},*_2), ..., (n_P_1,...,n_P_{ndims_P},*_P).
    indexing must be either 'ij' or 'xy'. Defaults to 'ij' like pytorch but unlike numpy.
    If ndims is None, meshes over all dimensions. ndims can be an int or an Iterable.
    Negative ndim has the same semantic meaning as slicing with a negative index.
    
    If indexing == 'ij', returns tensors of shapes 
    (n_1_1,...,n_1_{ndims_1},n_2_1,...,n_2_{ndims_2},...,n_P_1,...,n_P_{ndims_P},*_1),
    (n_1_1,...,n_1_{ndims_1},n_2_1,...,n_2_{ndims_2},...,n_P_1,...,n_P_{ndims_P},*_2),
    ...
    (n_1_1,...,n_1_{ndims_1},n_2_1,...,n_2_{ndims_2},...,n_P_1,...,n_P_{ndims_P},*_P)
    
    Otherwise, returns tensors of shapes
    (n_2_1,...,n_2_{ndims_2},n_1_1,...,n_1_{ndims_1},...,n_P_1,...,n_P_{ndims_P},*_1),
    (n_2_1,...,n_2_{ndims_2},n_1_1,...,n_1_{ndims_1},...,n_P_1,...,n_P_{ndims_P},*_2),
    ...
    (n_2_1,...,n_2_{ndims_2},n_1_1,...,n_1_{ndims_1},...,n_P_1,...,n_P_{ndims_P},*_P)
    
    IMPORTANT: Data is NOT copied just like pytorch, but unlike numpy which copies by default.
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")
    
    if not isinstance(ndims, collections.abc.Iterable):
        ndims = [ndims] * len(tensors)
    elif iter(ndims) is ndims:
        ndims = list(ndims)
        
    if not all(isinstance(ndim, int) or ndim is None for ndim in ndims):
        raise TypeError(f"ndim must be None or an int, but got {ndims}.")

    if len(ndims) != len(tensors):
        raise ValueError(
            f"""ndims must have same number of elements as input tensors,
            but got {len(ndims)=} and {len(tensors)=}."""
        )

    tensors = tuple(np.asanyarray(tensor) for tensor in tensors)
        
    if not all(
        abs(ndim) <= tensor.ndim
        for tensor, ndim in zip(tensors, ndims) if isinstance(ndim, int)
    ):
        raise ValueError(
            f"""abs(ndim) cannot be greater than the corresponding tensor.ndim, but
            {ndims=} while tensor ndims={[tensor.ndim for tensor in tensors]}."""
        )
    
    pre_shapes, post_shapes = zip(*(
        (tensor.shape[:ndim], () if ndim is None else tensor.shape[ndim:])
        for tensor, ndim in zip(tensors, ndims)
    ))
    shared_shape = tuple(itertools.chain(*pre_shapes))
    tensors = (
        np.broadcast_to(tensor.reshape(shape + post_shape), shared_shape + post_shape)
        for tensor, shape, post_shape in zip(tensors, meshshape(*pre_shapes), post_shapes)
    )
    
    if indexing == 'ij':
        return list(tensors)

    ndim0, ndim1 = len(pre_shapes[0]), len(pre_shapes[1])
    return [
        np.moveaxis(tensor, tuple(range(ndim0, ndim0 + ndim1)), tuple(range(ndim1)))
        for tensor in tensors
    ]

def inv_perm(p):
    """
    Code taken from: https://stackoverflow.com/questions/11649577/how-to-invert-a-permutation-array-in-numpy
    Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    p = np.asanyarray(p) # in case p is a tuple, etc.
    s = np.empty_like(p)
    s[p] = np.arange(p.size)
    return s

def unique_rows(arr, sorted=True, return_index=False, return_inverse=False, return_counts=False):
    """
    Code modified from https://github.com/scikit-image/scikit-image/blob/main/skimage/util/unique.py
    Much faster than np.unique(arr, axis=0)
    See https://github.com/numpy/numpy/issues/11136
    Set sorted=False for maximum speed.
    In this implementation, equal_nan=True.
    
    For speed comparison:
    N, M = 100000,100
    arr = np.random.randint(M, size=(N,2)).astype(float)
    
    %timeit unique_rows(arr, sorted=False, return_index=True, return_inverse=True)
    25.9 ms ± 386 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    
    %timeit np.unique(arr, return_index=True, return_inverse=True, axis=0)
    85.8 ms ± 240 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
    """
    if arr.ndim != 2:
        raise ValueError("unique_rows() only makes sense for 2D arrays, "
                         f"got {arr.ndim}")
    # the view in the next line only works if the array is C-contiguous
    arr = np.ascontiguousarray(arr)
    # np.unique() finds identical items in a raveled array. To make it
    # see each row as a single item, we create a view of each row as a
    # byte string of length itemsize times number of columns in `ar`
    arr_row_view = arr.view(f"|S{arr.itemsize * arr.shape[1]}")
    out = np.unique(arr_row_view, return_index=True, return_inverse=return_inverse, return_counts=return_counts)
    out = list(out)
    out[0] = arr[out[1]]
    
    if sorted:
        idx = np.lexsort(out[0].T[::-1])
        out[0] = out[0][idx]
        if return_index:
            out[1] = out[1][idx]
        if return_inverse:
            out[2] = inv_perm(idx)[out[2]]
        if return_counts:
            i = 3 if return_inverse else 2
            out[i] = out[i][idx]
    
    if not return_index:
        out.pop(1)

    if len(out) == 1:
        return out[0]
    return out

def intersect_rows(larr, rarr, return_indices=False, **kwargs):
    """
    intersect1d but for 2D arrays. Uses the same trick as unique_rows.
    """
    if larr.ndim != 2 or rarr.ndim != 2:
        raise ValueError("intersect_rows() only makes sense for 2D arrays, "
                         f"got {larr.ndim=}, {rarr.ndim=}")
    if larr.shape[1] != rarr.shape[1]:
        raise ValueError(f"larr and rarr must both have same number of columns, but {larr.shape[1]=}, {rarr.shape[1]=}.")
        
    N_cols = larr.shape[1]
    larr, rarr = np.ascontiguousarray(larr), np.ascontiguousarray(rarr)
    larr_row_view, rarr_row_view = larr.view(f"|S{larr.itemsize * N_cols}"), rarr.view(f"|S{rarr.itemsize * N_cols}")
    
    out = np.intersect1d(larr_row_view, rarr_row_view, return_indices=True, **kwargs)
    out = list(out)
    out[0] = larr[out[1]]
    
    if not return_indices:
        return out[0]
    return out
    
def repeat(arr, repeats, chunks=None):
    """
    Generalized np.repeat
    Copied from @MadPhysicist's solution: https://stackoverflow.com/questions/63510977/repeat-but-in-variable-sized-chunks-in-numpy
    
    Example due to @MadPhysicist in the same link:
    arr = np.array([0, 1, 2, 10, 11, 20, 21, 22, 23])
    #               >     <  >    <  >            <
    chunks = np.array([3, 2, 4])
    repeats = np.array([1, 3, 2])
    
    print(repeat(arr, repeats, chunks=chunks))
    >>>      [0, 1, 2, 10, 11, 10, 11, 10, 11, 20, 21, 22, 23, 20, 21, 22, 23])
    # repeats:>  1  <  >         3          <  >              2             <
    """
    if chunks is None:
        return np.repeat(arr, repeats)
    
    arr, repeats, chunks = np.asanyarray(arr), np.asanyarray(repeats), np.asanyarray(chunks)
    
    if arr.ndim != 1 or repeats.ndim != 1 or chunks.ndim != 1:
        raise ValueError(f"arr, repeats, and chunks must all be 1D, but {arr.ndim=}, {repeats.ndim=} and {chunks.ndim=}.")
    if len(repeats) != len(chunks):
        raise ValueError(f"repeats and chunks must have the same length, but {len(repeats)=} and {len(chunks)=}.")
    if chunks.sum() != len(arr):
        raise ValueError(f"sum of chunks must be the length of arr, but {chunks.sum()=} and {len(arr)=}.")

    regions = chunks * repeats
    index = np.arange(regions.sum())

    segments = np.repeat(chunks, repeats)
    resets = np.cumsum(segments[:-1])
    offsets = np.zeros_like(index)
    offsets[resets] = segments[:-1]
    offsets[np.cumsum(regions[:-1])] -= chunks[:-1]

    index -= np.cumsum(offsets)

    out = arr[index]
    
    return out



import functools
import logging
from datetime import datetime

import numpy as np
import numpy.ma as ma
import torch
import pandas as pd

from .sp import stats

logger = logging.getLogger(__name__)

def _formatter(x, float_format=None, datetime_format=None, verbose=False):
    if float_format is None:
        float_format = '{:.4f}'.format
    if datetime_format is None:
        datetime_format = '{:%Y-%m-%d %H:%M:%S}'.format
        
    if isinstance(x, np.ndarray):
        if x.ndim == 0:
            if ma.is_masked(x):
                return "NA"
            elif isinstance(x, np.inexact):
                return float_format(x)
            else:
                return str(x)
        if verbose:
            return f"np.ndarray {x.shape} ({x.dtype})"
        return f"{x.shape}"
    if isinstance(x, torch.Tensor):
        if x.ndim == 0:
            if x.is_floating_point():
                return float_format(x.item())
            else:
                return str(x.item())
        if verbose:
            return f"torch.Tensor {tuple(x.shape)} ({x.dtype})"
        return f"{tuple(x.shape)}"
    if isinstance(x, tuple):
        if verbose:
            return f"tuple ({len(x)}) {tuple([type(e).__name__ for e in x])}"
        return f"tuple ({len(x)})"
    if isinstance(x, list):
        if verbose:
            return f"list ({len(x)}) {[type(e).__name__ for e in x]}"
        return f"list ({len(x)})"
    if isinstance(x, dict):
        if verbose:
            d = {k: type(v).__name__ for k, v in x.items()}
            return f"dict ({len(x)}) {d}"
        return f"dict ({len(x)})"
    if isinstance(x, np.inexact):
        return float_format(x)
    if isinstance(x, datetime):
        return datetime_format(x)
    return str(x)

def display(df, float_format=None, datetime_format=None, verbose=False, **kwargs):
    default_kwargs = {
        'max_rows': 6,
        'show_dimensions': True,
        'formatters': {k: functools.partial(_formatter, float_format=float_format, datetime_format=datetime_format, verbose=verbose) for k in df.columns},
    }
    default_kwargs.update(kwargs)

    try:
        get_ipython
        from IPython.display import display as ipy_display
        from IPython.core.display import HTML
    except:
        raise RuntimeError("display only works in jupyter notebook")
        
    ipy_display(HTML(df.to_html(**default_kwargs)))
    
def digitize(x, column=None, colname=None, copy=True, new_cols=True, **kwargs):
    if isinstance(x, pd.Series):
        if column is not None:
            logger.warning("Attemping to digitize pd.Series with non-None column argument. column argument is ignored.")
        if colname is None:
            colname = 'x'
        column = colname
        x = pd.DataFrame({colname: x})
            
    elif isinstance(x, pd.DataFrame):
        if colname is not None:
            logger.warning("Attemping to digitize pd.DataFrame with non-None colname argument. colname argument is ignored.")
        if len(x.columns) == 1:
            column = x.columns.values[0] if column is None else column
        if len(x.columns) > 1 and column is None:
            raise ValueError("column must be provided if number of columns in the provided dataframe is greater than 1.")
            
    else:
        raise TypeError(f"x must be pd.Series or pd.DataFrame, but {type(x)} provided.")
        
    if copy:
        x = x.copy() # not necessary if x was a pd.Series, but whatever
        
    if isinstance(column, str):
        sample = revert_dtypes(x[column]).to_numpy()
        bin_nums, centers, edges = stats.bin(sample, nan_policy='omit', **kwargs)
        if new_cols:
            x[f'{column}_bin_center'] = centers[bin_nums]
            x[f'{column}_bin_ledge'] = np.array([-np.inf] + list(edges))[bin_nums]
            x[f'{column}_bin_redge'] = np.array(list(edges) + [np.inf])[bin_nums]
        else:
            x[column] = centers[bin_nums]
    else:
        sample = np.array([revert_dtypes(x[c]).to_numpy() for c in column]).T # (N,D)
        bin_nums, centers, edges = stats.bin_dd(sample, nan_policy='omit', **kwargs)
        for b, c, e, col in zip(bin_nums, centers, edges, column):
            if new_cols:
                x[f'{col}_bin_center'] = c[b]
                x[f'{col}_bin_ledge'] = np.array([-np.inf] + list(e))[b]
                x[f'{col}_bin_redge'] = np.array(list(e) + [np.inf])[b]
            else:
                x[col] = c[b]
            
    if new_cols:
        return x
    return x, edges

def mean(df, *, by, y='y'):
    df = df.copy()
    yerr = f'{y}err'
    group = df.groupby(by)
    df = pd.DataFrame({
        y: group[y].mean(),
        yerr: group[y].sem(),
    }).reset_index()
    return df

def meanerr(df, *, by, y='y', yerr='yerr'):
    assert all([x not in ['y_', 'yvar'] for x in [y, yerr]])
    df = df.copy()
    df['y_'] = df[y] + df[yerr]*0
    df['yvar'] = df[y]*0 + df[yerr]**2
    group = df.groupby(by)
    df = pd.DataFrame({
        y: group['y_'].mean(),
        yerr: (group['yvar'].mean() + group['y_'].var())**0.5 / group['y_'].count()**0.5,
    }).reset_index()
    return df
    
def get_np_dtype(x):
    if not isinstance(x, pd.Series):
        raise TypeError(f'input must be pd.Series, but got {type(x)}')
    
    dtype = str(x.dtype)
    if 'Int' in dtype:
        if x.isnull().any():
            return 'float'
        return dtype.lower()
    elif 'Float' in dtype:
        return dtype.lower()
    elif 'boolean' in dtype:
        if x.isnull().any():
            return 'float'
        return 'bool'
    elif pd.api.types.is_numeric_dtype(dtype): # could be complex, for example
        return dtype
    else:
        return 'object'
    
def revert_dtypes(x):
    if isinstance(x, pd.Series):
        return x.astype(get_np_dtype(x))
    
    if isinstance(x, pd.DataFrame):
        dtypes = {col: get_np_dtype(x[col]) for col in x.columns}
        return x.astype(dtypes)
    
    raise TypeError(f'x must be pd.Series or pd.DataFrame, but got {type(x)}')

def to_tensor(x):
    is_list = False
    if isinstance(x, list):
        x = np.array(x)
        is_list = True
    if isinstance(x, np.ndarray):
        try:
            if x.dtype.kind in ['m', 'M', 'O', 'S', 'U', 'V']:
                return x.tolist() if is_list else x
            if x.dtype.kind == 'f':
                return torch.as_tensor(x, dtype=torch.float32)
            return torch.as_tensor(x)
        except Exception as err:
            raise RuntimeError(f'Error converting to tensor: {x.dtype=}, {x=}') from err
    return x

# update: df.join actually seems faster
# def cross_join(*dfs, maintain_dtypes=True):
#     """
#     Efficient cross/cartesian product of numeric dataframes. Should be faster than df.join()
#     If maintain_dtypes=True, will ensure that the resulting DataFrame has the same dtypes as the original DataFrames.
#     However, maintain_dtypes=True is very slow in general.
#     """
#     columns = [column for df in dfs for column in df.columns]
#     dtypes = {k: v for df in dfs for k, v in df.dtypes.items()}
#     dfs = meshgrid_dd(*(revert_dtypes(df).to_numpy() for df in dfs)) # reverting dtypes makes meshgrid faster if all columns are numeric
#     df = np.concatenate([df.reshape(-1,df.shape[-1]) for df in dfs], axis=-1)
#     df = pd.DataFrame(df, columns=columns)
#     if maintain_dtypes:
#         return df.astype(dtypes)
#     return df


import logging
import contextlib

from . import argparse

def basic_config(level='INFO', format='%(levelname)s: %(name)s: %(message)s', capture_warnings=True):
    logging.basicConfig(level=getattr(logging, level), format=format)
    logging.captureWarnings(capture_warnings)
    
def basic_parser(**kwargs):
    parser = argparse.default_parser(basic_config, configs={
        'level': dict(opt_str=['--ll', '--log-level'], choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']),
        'format': dict(opt_str=['--lf', '--log-format']),
    }, **kwargs)
    return parser

def get_logger(name=None):
    return logging.getLogger(name)

@contextlib.contextmanager
def set_level(logger, level):
    current_level = logger.level
    try:
        logger.setLevel(level)
        yield
    finally:
        logger.setLevel(current_level)


import collections

def pformat(d, indent=0, spaces=4, depth=-1, formatters=None, verbose=True):
    output = ''
    for key, value in d.items():
        output += f"{' ' * spaces * indent}'{str(key)}':\n"
        if depth != 0 and isinstance(value, dict):
            output += f"{pformat(value, indent=indent+1, spaces=spaces, depth=depth - 1, formatters=formatters, verbose=verbose)}\n"
        else:
            if formatters and type(value) in formatters:
                value = formatters[type(value)](value)
            elif verbose:
                value = str(value)
            else:
                value = str(type(value))
            output += '\n'.join([f"{' ' * spaces * (indent+1)}{line}" for line in value.split('\n')]) + '\n'
    return output.rstrip('\n')

def pprint(d, **kwargs):
    print(pformat(d, **kwargs))

def iter_str(l):
    if isinstance(l, str):
        return repr(l)
        
    if not isinstance(l, (tuple, list, set, dict)):
        return str(l)

    if isinstance(l, dict):
        s = ', '.join(f'{repr(k)}: {v}' for k, v in l.items())
        return '{' + s + '}'
        
    s = ', '.join(iter_str(v) for v in l)

    if isinstance(l, set):
        return '{' + s + '}'

    if isinstance(l, tuple):
        return '(' + s + ')'
    
    return '[' + s + ']'
    
def pformat_english(*args):
    """
    Formats arguments in a way that follows English grammatical rules.
    """
    length = len(args)
    
    if length == 0:
        return ""
    
    if length == 1:
        return str(args[0])
    
    if length == 2:
        return f'{args[0]} and {args[1]}'
    
    return f"{', '.join(map(str, args[:-1]))}, and {args[-1]}"


import traceback
import warnings
import sys
import contextlib

def _warn_with_traceback(message, category, filename, lineno, file=None, line=None):

    log = file if hasattr(file, 'write') else sys.stderr
    traceback.print_stack(file=log)
    log.write(warnings.formatwarning(message, category, filename, lineno, line))

@contextlib.contextmanager
def warn_with_traceback():
    showwarning = warnings.showwarning
    try:
        warnings.showwarning = _warn_with_traceback
        yield
    finally:
        warnings.showwarning = showwarning


import json
import copy

from . import exceptions

class Parameter:
    def __init__(self, data=None, dtype=None, is_jsonable=True, deepcopy=False, load_callback=None):
        if dtype is not None:
            self.dtype = dtype
            
            if data is None:
                try:
                    data = dtype()
                except TypeError as err:
                    raise TypeError("dtype must be a python class that can be constructed without arguments")
        else:
            if data is None:
                raise ValueError("dtype must not be None when data is None")
                
            self.dtype = type(data)
            
        if is_jsonable:
            try:
                json.dumps(data)
            except (TypeError, OverflowError):
                raise exceptions.InvalidConfigParameter(f"data {data} is not json serializable," 
                                                        " try converting it to a compatible object type."
                                                        " If not possible, you can set is_jsonable=False,"
                                                        " but then you can't save the config dict as json.") from None
        self.deepcopy = deepcopy
        self.load_callback = load_callback
        self.set_data(data)
        
    def set_data(self, data):
        if self.deepcopy:
            data = copy.deepcopy(data)
        self._data = self.dtype(data)  # shallow copy
        
    def load(self, data):
        self.set_data(data)
        if self.load_callback is not None:
            self.load_callback(self)
        
    @property
    def data(self):
        data = self.dtype(self._data) # shallow copy
        if self.deepcopy:
            data = copy.deepcopy(data)
        return data
    
    def __repr__(self):
        return f"Parameter(data={self.data}, dtype={self.dtype}, deepcopy={self.deepcopy})"
        
class Configurable:
    # define __init__ with super() call to make this class suitable for multiple inheritance
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__parameters = {}
        self.__configurables = {}
        
    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if isinstance(value, Parameter):
            return value.data
        return value
    
    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if isinstance(value, Parameter):
            if hasattr(self, '_Configurable__parameters'):
                self.__parameters[name] = value
            else:
                raise RuntimeError("Configurable must be initalized before setting Parameter")
        elif isinstance(value, Configurable):
            if hasattr(self, '_Configurable__configurables'):
                self.__configurables[name] = value
            else:
                raise RuntimeError("Configurable must be initalized before setting Configurable")
        else:
            if hasattr(self, '_Configurable__parameters') and name in self.__parameters:
                del self.__parameters[name]
            elif hasattr(self, '_Configurable__configurables') and name in self.__configurables:
                del self.__configurables[name]
    
    def __delattr__(self, name):
        if hasattr(self, '_Configurable__parameters') and name in self.__parameters:
            del self.__parameters[name]
        elif hasattr(self, '_Configurable__configurables') and name in self.__configurables:
            del self.__configurables[name]
        super().__delattr__(name)
    
    @property
    def params_dict(self):
        return self.__parameters
    
    @property
    def configurables_dict(self):
        return self.__configurables
    
    def config_dict(self):
        """
        Returns a configuration dictionary. Uses a flat instead of hierarchical structure
        because otherwise config parameters that are dictionaries will be ambiguous when trying
        to load them back in.
        """
        d = {}
        for name, param in self.params_dict.items():
            d[name] = param.data
        for name, configurable in self.configurables_dict.items():
            sub_d = configurable.config_dict()
            d.update({f'{name}.{sub_key}': sub_value for sub_key, sub_value in sub_d.items()})
        return d
    
    def _load_config_dict(self, config_dict, prefix, missing_keys, unexpected_keys):
        for name, param in self.params_dict.items():
            key = prefix + name
            try:
                value = config_dict[key]
                param.load(value)
            except KeyError as err:
                missing_keys.append(key)
                
            
        for key in config_dict.keys():
            if key.startswith(prefix):
                key_parts = key[len(prefix):].split('.')
                if len(key_parts) == 1 and key_parts[0] not in self.__parameters:
                    unexpected_keys.append(key)
    
    def load_config_dict(self, config_dict, strict=True):
        missing_keys = []
        unexpected_keys = []
        
        def load(configurable, prefix=''):
            configurable._load_config_dict(config_dict, prefix, missing_keys, unexpected_keys)
            for name, sub_configurable in configurable.configurables_dict.items():
                load(sub_configurable, prefix + name + '.')
        
        load(self)
        
        if strict:
            if not (len(missing_keys) == 0 and len(unexpected_keys) == 0):
                print(missing_keys)
                print(unexpected_keys)
                raise AssertionError("strict=True, but there is mismatch when loading config dict.")
        
# class Configurable:
#     # define __init__ with super() call to make this class suitable for multiple inheritance
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self._configs = {}
        
#     @property
#     def configs(self):
#         return self._configs
    
#     def __getattr__(self, name):
#         """
#         First checks whether name can be found using the super class's __getattr__
#         If not, then check whether name is in self._configs
#         """
#         try:
#             return super().__getattr__(name)
#         except AttributeError as err:
#             # Very important check! if self._configs doesn't exist, this will indefinitely recurse.
#             # This is because self._configs has already failed, so we shouldn't try to call self._configs again
#             # in the exception handling block.
#             # In general, for any name such self.name is called in this block (whether this is called explicitly
#             # or implicitly), we need to raise an error immediately whenever we encounter such name to prevent recursion.
#             if name == '_configs':
#                 raise
#             return self._configs[name]

#     def __setattr__(self, name, value):
#         if isinstance(value, Parameter):
#             if hasattr(self, name):
#                 delattr(self, name)
#             self.configs[name] = value.data
#         else:
#             if name in self.configs:
#                 del self.configs[name]
#             super().__setattr__(name, value)


import datetime

import pytest

import hyclib as lib

@pytest.mark.parametrize(
    'td, fmt, expected',
    [
        (
            datetime.timedelta(days=2, hours=1, minutes=5, seconds=12),
            '%Hh%Ss',
            '49h312s',
        ),
        (
            datetime.timedelta(weeks=2, days=4, hours=100, minutes=5, seconds=12, milliseconds=150, microseconds=5),
            '%w weeks %d days %H:%M:%S.%f',
            '3 weeks 1 days 04:05:12.150005',
        ),
        (
            datetime.timedelta(days=-4, hours=10, minutes=5, seconds=12, milliseconds=150, microseconds=5),
            '%d days, %H:%M:%S.%f',
            None,
        ),
    ],
)      
def test_strftime(td, fmt, expected):
    if expected is None:
        expected = str(td)
        
    output = lib.datetime.strftime(td, fmt)

    assert output == expected


import collections
import contextlib
import itertools
import warnings

import torch

from hyclib.np.core import meshshape

__all__ = ['isconst', 'inv_perm', 'lexsort', 'unique', 'meshgrid', 'use_deterministic_algorithms']

def isconst(x, dim=None, **kwargs):
    x = torch.as_tensor(x)
    
    if dim is None:
        x = x.reshape(-1)
    else:
        if isinstance(dim, int):
            dim = [dim]
        if len(dim) == 0:
            return torch.ones(x.shape, device=x.device, dtype=torch.bool)
        dim = sorted([d % x.ndim for d in dim])[::-1]
        for d in dim:
            x = torch.movedim(x, d,-1)
        x = x.reshape(*x.shape[:-len(dim)],-1)
        
    if x.is_floating_point():
        return torch.isclose(x[...,:-1], x[...,1:], **kwargs).all(dim=-1)
    return (x[...,:-1] == x[...,1:]).all(dim=-1)

def inv_perm(x):
    x_inv = torch.empty_like(x)
    x_inv[x] = torch.arange(len(x), device=x.device)
    return x_inv

def lexsort(keys, dim=-1):
    keys = keys if isinstance(keys, torch.Tensor) else torch.stack(keys)
    if keys.ndim < 2:
        raise ValueError(f"keys must be at least 2 dimensional, but {keys.ndim=}.")
    if len(keys) == 0:
        raise ValueError(f"Must have at least 1 key, but {len(keys)=}.")
    
    # to understand reasoning, see the code at
    # https://github.com/wouterkool/attention-learn-to-route/blob/master/utils/lexsort.py#L39-L48
    idx = keys[0].argsort(dim=dim, stable=True)
    for k in keys[1:]:
        idx = idx.gather(dim, k.gather(dim, idx).argsort(dim=dim, stable=True))
    
    return idx

def _unique_sorted(x, dim=None, return_index=False, return_inverse=False, return_counts=False, first_index=True):
    if dim is None:
        x = x.reshape(-1)
        dim = 0
        
    # reshape to a 2D tensor where we compute unique rows
    x = x.movedim(dim, 0)
    shape = x.shape
    x = x.reshape(shape[0], -1)

    sort_idx = lexsort(x.t().flip(0))  # requires torch>=2.1.0 for this to work correctly on MPS
    
    x = x[sort_idx]
    
    if return_index:
        x, inverse, counts = torch.unique_consecutive(x, dim=0, return_inverse=True, return_counts=True)
        tot_counts = torch.cat((counts.new_zeros(1), counts.int().cumsum(dim=0).long()))[:-1] # cast to int32 due to MPS limitation
        index = torch.arange(len(inverse), device=tot_counts.device)[tot_counts]
        
        out = {'x': x, 'index': index}
        if return_inverse:
            out['inverse'] = inverse
        if return_counts:
            out['counts'] = counts

        # the following code doesn't work right now on MPS due to MPS determinism bug
        # ret = torch.unique_consecutive(x, dim=0, return_inverse=True, return_counts=return_counts)

        # x, inverse = ret[0], ret[1]
        # args = (len(inverse)-1, -1, -1) if first_index else (len(inverse),)
        # with use_deterministic_algorithms() if first_index else contextlib.nullcontext():
        #     # scatter_ is non-deterministic by default and would result in a random index rather than the first index
        #     index = inverse.new_empty(len(x)).scatter_(
        #         0,
        #         inverse.flip(0) if first_index else inverse,
        #         torch.arange(*args, dtype=inverse.dtype, device=inverse.device),
        #     )
        
        # # gather results into a dictionary
        # out = {'x': x, 'index': index}
        # if return_inverse:
        #     out['inverse'] = inverse
        # if return_counts:
        #     out['counts'] = ret[2]

    else:
        ret = torch.unique_consecutive(x, dim=0, return_inverse=return_inverse, return_counts=return_counts)

        # gather results into a dictionary
        ret = list(ret) if isinstance(ret, tuple) else [ret]
        out = {'x': ret.pop(0)}
        if return_inverse:
            out['inverse'] = ret.pop(0)
        if return_counts:
            out['counts'] = ret.pop(0)
            
    # correct for the fact that we reshaped and sorted before computing unqiue
    out['x'] = out['x'].reshape(-1, *shape[1:]).movedim(0, dim)
    if return_index or return_inverse:
        if return_index:
            out['index'] = sort_idx[out['index']]
        if return_inverse:
            sort_idx_inv = inv_perm(sort_idx)
            out['inverse'] = out['inverse'][sort_idx_inv]
    
    if len(out) == 1:
        return out['x']
    return tuple(out.values())

def unique(x, dim=None, sorted=True, return_index=False, return_inverse=False, return_counts=False, equal_nan=False, first_index=True):
    """
    Note:
     - When sorted=False, there is no guarantee on what the result order will be.
     - The result of this function when sorted=True, return_index=False is slightly different from numpy
     with return_index=False since numpy does not use stable sort when return_index=False, only when return_index=True.
     - If return_index=True and first_index=True, the returned index is the first index of the unique element. Otherwise,
        the returned index can correspond to any index of the unique element, and the behavior is non-deterministic.
    """
    
    if equal_nan:
        raise NotImplementedError("pytorch currently does not support equal_nan=True")
    
    # unique_dim currently not implemented for MPS, so do _unique_sorted instead
    # which uses unique_consecutive (which has MPS implementation) instead of unique_dim.
    # the reason there is no MPS unqiue_dim support right now is that there is currently 
    # no native MPS support for sorting with custom compare and equal functions, so one 
    # has to write a custom MPS Kernel which means it might take a long while before this gets implemented.
    if sorted or x.device.type == 'mps':
        return _unique_sorted(x, dim=dim, return_index=return_index, return_inverse=return_inverse, return_counts=return_counts)
    
    if dim is None:
        x = x.reshape(-1)
        dim = 0
    
    # reshape to a 2D tensor where we compute unique rows
    x = x.movedim(dim, 0)
    shape = x.shape
    x = x.reshape(shape[0], -1)
    
    # separate rows with nans from rows without nans and only compute unique on rows without nans.
    # This avoids torch.unique bug when dim is not None and input has nan.
    isnan = x.isnan().any(dim=1)
    if not isnan.any():
        x_not_isnan, x_isnan = x, torch.empty((0, x.shape[1]), dtype=x.dtype, device=x.device) # slighty optimization to prevent large copy
    else:
        x_not_isnan, x_isnan = x[~isnan], x[isnan]
    N_not_isnan = len(x_not_isnan)

    if return_index:
        # # code copied from https://github.com/pytorch/pytorch/issues/36748#issuecomment-1474368922
        # x_not_isnan, inverse, counts = torch.unique(x_not_isnan, dim=0, return_inverse=True, return_counts=True)
        # inv_sort_idx = inverse.argsort(stable=True)
        # tot_counts = torch.cat((counts.new_zeros(1), counts.cumsum(dim=0)))[:-1]
        # index = inv_sort_idx[tot_counts]

        # out = {'x': x_not_isnan, 'index': index}
        # if return_inverse:
        #     out['inverse'] = inverse
        # if return_counts:
        #     out['counts'] = counts

        ret = torch.unique(x_not_isnan, dim=0, return_inverse=True, return_counts=return_counts)

        x_not_isnan, inverse = ret[0], ret[1]
        args = (len(inverse)-1, -1, -1) if first_index else (len(inverse),)
        with use_deterministic_algorithms() if first_index else contextlib.nullcontext():
            # scatter_ is non-deterministic by default and would result in a random index rather than the first index
            index = inverse.new_empty(len(x_not_isnan)).scatter_(
                0,
                inverse.flip(0) if first_index else inverse,
                torch.arange(*args, dtype=inverse.dtype, device=inverse.device),
            )

        # gather results into a dictionary
        out = {'x': x_not_isnan, 'index': index}
        if return_inverse:
            out['inverse'] = inverse
        if return_counts:
            out['counts'] = ret[2]

    else:
        ret = torch.unique(x_not_isnan, dim=0, return_inverse=return_inverse, return_counts=return_counts)
    
        # gather results into a dictionary
        ret = list(ret) if isinstance(ret, tuple) else [ret]
        out = {'x': ret.pop(0)}
        if return_inverse:
            out['inverse'] = ret.pop(0)
        if return_counts:
            out['counts'] = ret.pop(0)
        
    # append nans to the end of the unique tensor and handle index, inverse, and counts accordingly
    N_unique_not_isnan = len(out['x'])
    out['x'] = torch.cat([out['x'], x_isnan])
    if return_index:
        out['index'] = torch.cat([out['index'], torch.arange(len(x_isnan), device=x.device) + N_not_isnan])
    if return_inverse:
        out['inverse'] = torch.cat([out['inverse'], torch.arange(len(x_isnan), device=x.device) + N_unique_not_isnan])
    if return_counts:
        out['counts'] = torch.cat([out['counts'], out['counts'].new_ones(len(x_isnan))])
    
    # correct for the fact that we reshaped and modified element order before computing unqiue
    out['x'] = out['x'].reshape(-1, *shape[1:]).movedim(0, dim)
    if return_index or return_inverse:
        sort_idx = torch.cat(((~isnan).nonzero(), isnan.nonzero())).squeeze()
        if return_index:
            out['index'] = sort_idx[out['index']]
        if return_inverse:
            sort_idx_inv = inv_perm(sort_idx)
            out['inverse'] = out['inverse'][sort_idx_inv]
    
    if len(out) == 1:
        return out['x']
    return tuple(out.values())

def meshgrid(*tensors, indexing='ij', dims=None, sparse=False):
    """
    The most general version of torch.meshgrid. Meshes together tensors of shapes
    (*00, n0[dims[0][0]], ..., n0[dims[0][1]], *01),
    (*10, n1[dims[1][0]], ..., n1[dims[1][1]], *11),
    ..., 
    (*P0, nP[dims[P][0]], ..., nP[dims[P][1]], *P1),
    indexing must be either 'ij' or 'xy'. Defaults to 'ij' like pytorch but unlike numpy.
    If dims is None, meshes over all dimensions. dims can be None, an int,
    or a length-P list/tuple of length-2 list/tuple of ints or None, where P = len(tensors).
    Negative dim has the same semantic meaning as slicing with a negative index.
    If sparse, new dimensions are of length one instead of the broadcasted length.
    
    If indexing == 'ij', returns tensors of shapes 
    (*00, n0[dims[0][0]], ..., n0[dims[0][1]], n1[dims[1][0]], ..., n1[dims[1][1]], ..., nP[dims[P][0]], ..., nP[dims[P][1]], *01),
    (*10, n0[dims[0][0]], ..., n0[dims[0][1]], n1[dims[1][0]], ..., n1[dims[1][1]], ..., nP[dims[P][0]], ..., nP[dims[P][1]], *11),
    ...,
    (*P0, n0[dims[0][0]], ..., n0[dims[0][1]], n1[dims[1][0]], ..., n1[dims[1][1]], ..., nP[dims[P][0]], ..., nP[dims[P][1]], *P1),
    
    Otherwise, returns tensors of shapes
    (*00, n1[dims[1][0]], ..., n1[dims[1][1]], n0[dims[0][0]], ..., n0[dims[0][1]], ..., nP[dims[P][0]], ..., nP[dims[P][1]], *01),
    (*10, n1[dims[1][0]], ..., n1[dims[1][1]], n0[dims[0][0]], ..., n0[dims[0][1]], ..., nP[dims[P][0]], ..., nP[dims[P][1]], *11),
    ...,
    (*P0, n1[dims[1][0]], ..., n1[dims[1][1]], n0[dims[0][0]], ..., n0[dims[0][1]], ..., nP[dims[P][0]], ..., nP[dims[P][1]], *P1),
    
    IMPORTANT: Data is NOT copied just like pytorch, unlike numpy which copies by default.
    """
    if not indexing in {'ij', 'xy'}:
        raise ValueError(f"indexing must 'ij' or 'xy', but got {indexing}.")
    
    if dims is None or isinstance(dims, int):
        dims = [(None, dims)] * len(tensors)
    elif not (
        isinstance(dims, (list, tuple)) and len(dims) == len(tensors) and 
        all(
            (
                isinstance(dim, (list, tuple)) and len(dim) == 2 and 
                all(d is None or isinstance(d, int) for d in dim)
            )
            for dim in dims
        )
    ):
        raise TypeError(
            f"""dims must be None, an int, or a length-{len(tensors)}
            list/tuple of length-2 list/tuple of None or int, but {dims=}."""
        )

    pre_shapes, shapes, post_shapes = zip(*(
        (
            () if dim[0] is None else tensor.shape[:dim[0]],
            tensor.shape[dim[0]:dim[1]],
            () if dim[1] is None else tensor.shape[dim[1]:],
        )
        for tensor, dim in zip(tensors, dims)
    ))
    shapes = list(meshshape(*shapes, indexing=indexing))
    tensors = (
        tensor.reshape(pre_shape + shape + post_shape)
        for tensor, shape, pre_shape, post_shape in zip(tensors, shapes, pre_shapes, post_shapes)
    )
    if not sparse:
        shape = torch.broadcast_shapes(*shapes)
        tensors = (
            tensor.broadcast_to(pre_shape + shape + post_shape)
            for tensor, pre_shape, post_shape in zip(tensors, pre_shapes, post_shapes)
        )
    
    return list(tensors)

@contextlib.contextmanager
def use_deterministic_algorithms():
    """
    Context manager to use deterministic algorithms.
    """
    is_deterministic = torch.are_deterministic_algorithms_enabled()
    try:
        torch.use_deterministic_algorithms(True)
        yield
    finally:
        torch.use_deterministic_algorithms(is_deterministic)

##########################################
### Implemented but untested functions ###
##########################################

# import numpy as np

# def ravel_multi_index(multi_index, dims):
#     """
#     Similar to np.ravel_multi_index
#     """
#     multi_index = torch.stack(multi_index) if not isinstance(multi_index, torch.Tensor) else multi_index
#     dims = torch.as_tensor(dims, device=multi_index.device)
    
#     if len(multi_index) != len(dims):
#         raise ValueError(f"multi_index and dims must have same length, but {len(multi_index)=} and {len(dims)=}.")
    
#     if torch.is_floating_point(multi_index):
#         raise TypeError(f"multi_index must be integer dtype, but {multi_index.dtype=}.")
        
#     if (multi_index.min(dim=1).values < 0).any():
#         raise ValueError(f"multi_index must be non-negative, but {multi_index.min(dim=1).values=}.")
        
#     if (multi_index.max(dim=1).values >= dims).any():
#         raise ValueError(f"multi_index must be less than dims along each dimension, but {multi_index.max(dim=1).values=} and {dims=}.")
        
#     multipliers = np.cumprod((dims[1:].tolist() + [1])[::-1])[::-1]
#     return torch.stack([index * multiplier for index, multiplier in zip(multi_index, multipliers)], dim=0).sum(dim=0)

# def unravel_index(indices, shape, *, as_tuple=True):
#     r"""
#     Modified from https://github.com/pytorch/pytorch/pull/66687/files
    
#     Converts a `Tensor` of flat indices into a `Tensor` of coordinates for the given target shape.
#     Args:
#         indices: An integral `Tensor` containing flattened indices of a `Tensor` of dimension `shape`.
#         shape: The shape (can be an `int`, a `Sequence` or a `Tensor`) of the `Tensor` for which
#                the flattened `indices` are unraveled.
#     Keyword Args:
#         as_tuple: A boolean value, which if `True` will return the result as tuple of Tensors,
#                   else a `Tensor` will be returned. Default: `True`
#     Returns:
#         unraveled coordinates from the given `indices` and `shape`. See description of `as_tuple` for
#         returning a `tuple`.
#     .. note:: The default behaviour of this function is analogous to
#               `numpy.unravel_index <https://numpy.org/doc/stable/reference/generated/numpy.unravel_index.html>`_.
#     Example::
#         >>> indices = torch.tensor([22, 41, 37])
#         >>> shape = (7, 6)
#         >>> torch.unravel_index(indices, shape)
#         (tensor([3, 6, 6]), tensor([4, 5, 1]))
#         >>> torch.unravel_index(indices, shape, as_tuple=False)
#         tensor([[3, 4],
#                 [6, 5],
#                 [6, 1]])
#         >>> indices = torch.tensor([3, 10, 12])
#         >>> shape_ = (4, 2, 3)
#         >>> torch.unravel_index(indices, shape_)
#         (tensor([0, 1, 2]), tensor([1, 1, 0]), tensor([0, 1, 0]))
#         >>> torch.unravel_index(indices, shape_, as_tuple=False)
#         tensor([[0, 1, 0],
#                 [1, 1, 1],
#                 [2, 0, 0]])
#     """
#     def _helper_type_check(inp, name):
#         # `indices` is expected to be a tensor, while `shape` can be a sequence/int/tensor
#         if name == "shape" and isinstance(inp, collections.abc.Sequence):
#             for dim in inp:
#                 if not isinstance(dim, int):
#                     raise TypeError("Expected shape to have only integral elements.")
#                 if dim < 0:
#                     raise ValueError("Negative values in shape are not allowed.")
#         elif name == "shape" and isinstance(inp, int):
#             if inp < 0:
#                 raise ValueError("Negative values in shape are not allowed.")
#         elif isinstance(inp, torch.Tensor):
#             if torch.is_floating_point(inp):
#                 raise TypeError(f"Expected {name} to be an integral tensor, but found dtype: {inp.dtype}")
#             if torch.any(inp < 0):
#                 raise ValueError(f"Negative values in {name} are not allowed.")
#         else:
#             allowed_types = "Sequence/Scalar (int)/Tensor" if name == "shape" else "Tensor"
#             msg = f"{name} should either be a {allowed_types}, but found {type(inp)}"
#             raise TypeError(msg)

#     _helper_type_check(indices, "indices")
#     _helper_type_check(shape, "shape")

#     # Convert to a tensor, with the same properties as that of indices
#     if isinstance(shape, collections.abc.Sequence):
#         shape_tensor = indices.new_tensor(shape)
#     elif isinstance(shape, int) or (isinstance(shape, Tensor) and shape.ndim == 0):
#         shape_tensor = indices.new_tensor((shape,))
#     else:
#         shape_tensor = shape

#     # By this time, shape tensor will have dim = 1 if it was passed as scalar (see if-elif above)
#     assert shape_tensor.ndim == 1, "Expected dimension of shape tensor to be <= 1, "
#     f"but got the tensor with dim: {shape_tensor.ndim}."

#     # In case no indices passed, return an empty tensor with number of elements = shape.numel()
#     if indices.numel() == 0:
#         # If both indices.numel() == 0 and shape.numel() == 0, short-circuit to return itself
#         shape_numel = shape_tensor.numel()
#         if shape_numel == 0:
#             raise ValueError("Got indices and shape as empty tensors, expected non-empty tensors.")
#         else:
#             output = [indices.new_tensor([]) for _ in range(shape_numel)]
#             return tuple(output) if as_tuple else torch.stack(output, dim=1)

#     if torch.max(indices) >= torch.prod(shape_tensor):
#         raise ValueError("Target shape should cover all source indices.")

#     coefs = shape_tensor[1:].flipud().cumprod(dim=0).flipud()
#     coefs = torch.cat((coefs, coefs.new_tensor((1,))), dim=0)
#     coords = torch.div(indices[..., None], coefs, rounding_mode='trunc') % shape_tensor

#     if as_tuple:
#         return tuple(coords[..., i] for i in range(coords.size(-1)))
#     return coords